import numpy as np
import re


#workflow
# 1) a)  pass expression 


#expect that numpy is imported as np
#expression the formula you want to get the corresponding function
#coefficients a dictionnary: keys are the coefficients name in the expression an values are your values
#variable names the names of the variable you want to evaluate this function on
=======

def generate_specific_function(expression, coefficients, variable_names):
    # Define the function to only vary with specified variables
    def specific_function(*variable_values):
        local_vars = coefficients.copy()  # Copy the dictionary of fixed variables
        # Update variable names with values dynamically from the input
        local_vars.update(dict(zip(variable_names, variable_values)))
        local_vars['np'] = np  # Ensure np is available for np.log10 and other operations
        return eval(expression, {}, local_vars)
    # The number of inputs is now the number of variable names provided
    return np.frompyfunc(specific_function, len(variable_names), 1)


def extract_coef_names(expression):
    # Regex to find isolated letters (considered variables)
    pattern = r'\b[a-zA-Z]\b'
    # Find all matches and return as a set to remove duplicates
    matches = set(re.findall(pattern, expression))
    return sorted(matches) 


def create_coef_dict (coef_names, coef_values):
    if len(coef_names)==len(coef_values):
       out = dict(zip(coef_names, coef_values))
    else :
        raise TypeError("length of coefficietns names different from the length of coefficients values")
    return(out)

=======

def extract_variables_names(expression, suffix='_range'):
    # Construct the regex pattern dynamically to find words ending with the given suffix
    pattern = rf'\b\w+{re.escape(suffix)}\b'  # Use re.escape to safely include the suffix in the regex
    # Find all matches and return as a set to remove duplicates
    matches = set(re.findall(pattern, expression))
    return sorted(list(matches))  # Return a sorted list of unique matches



import logging
from pathlib import Path
from . import _setup_cfg

_logger = logging.getLogger('Innate')

REF_entry = _setup_cfg['parameter_labels']['param']

try:
    from astropy.io import fits
    astropy_check = True
except ImportError:
    astropy_check = False

try:
    import h5netcdf
    h5netcdf_check = True
except ImportError:
    h5netcdf_check = False


# Define library error function
class InnateError(Exception):
    """InnateError exception function"""


def load_dataset(fname: str):
    """
        Load the data grids and configuration from a digital file.

        Parameters
        ----------
        fname : str
            The file address or path of the file.

        Returns
        -------

        A dictionary with the data arrays, a dictionary with the global grids configuration and a dictionary with the local grid configuration.
            - For '.fits' files: (array_dict, common_cfg, local_cfg)
            - For '.nc' files: (array_dict, common_cfg, local_cfg)

        Raises
        ------
        InnateError
            If no dataset file is found at the specified location.
        KeyError
            If the file extension is not recognized.

        Notes
        -----
        This function checks the file extension and loads the dataset using the appropriate method:
        - '.fits' files are loaded using `fits_file_load`.
        - '.nc' files are loaded using `h5netcdf_file_load`.

        Examples
        --------
        >>> data = load_dataset('data/file.fits')
        >>> data = load_dataset('data/file.nc')
        """

    # Check the file location
    if type(fname).__name__ != 'UploadedFile':
        log_path = Path(fname)
        if not log_path.is_file():
            raise InnateError(f'Not dataset file found at {fname}')

        fname, ext = log_path, log_path.suffix

    else:
        fname, ext = fname, 'UploadedFile'

    # Run the appropriate file saving workflow
    match ext:

        case '.fits':
            return fits_file_load(fname)

        case '.nc':
            return h5netcdf_file_load(fname)

        case 'UploadedFile':
            return h5netcdf_file_load(fname)

        case _:
            return KeyError(f'The extension "{ext}" is not recognized, please use ".nc" or ".fits"')


def save_dataset(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    """
    Save the data grids and configuration to a digital file.

    Parameters
    ----------
    fname : str or pathlib.Path
        The name of the file where the dataset will be saved.
    grid_dict : dict
        Dictionary containing the data grid to be saved.
    common_cfg : dict
        Dictionary containing common configuration parameters.
    custom_cfg : dict
        Dictionary containing custom configuration parameters.

    Returns
    -------
    None

    Raises
    ------
    KeyError
        If the file extension is not recognized.

    Notes
    -----
    This function checks the file extension and saves the dataset using the appropriate method:
    - '.fits' files are saved using `fits_file_save`.
    - '.nc' files are saved using `h5netcdf_file_save`.

    Examples
    --------
    >>> save_dataset('data/output.fits', grid_dict, common_cfg, custom_cfg)
    >>> save_dataset('data/output.nc', grid_dict, common_cfg, custom_cfg)
    """

    # Check the file location
    fname = Path(fname)
    ext = fname.suffix

    # Check there is input data
    if len(grid_dict) == 0:
        _logger.warning(f'Output dataset does not contain any grid entry')

    # Run the appropriate file saving workflow
    match ext:

        case '.fits':
            return fits_file_save(fname, grid_dict, common_cfg, custom_cfg)

        case '.nc':
            return h5netcdf_file_save(fname, grid_dict, common_cfg, custom_cfg)

        case _:
            return KeyError(f'The extension "{ext}" is not recognized, please use ".nc" or ".fits"')


def fits_file_save(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    if astropy_check:

        # Create a primary HDU
        hdu_list = fits.HDUList([fits.PrimaryHDU()])

        # Generate the fits file
        for key, grid in grid_dict.items():
            hdu_i = fits.ImageHDU(grid, name=key)
            hdu_list.append(hdu_i)

        # Write the fits file
        hdu_list.writeto(fname, overwrite=True)

    else:
        raise InnateError(f'To open ".fits" files (astropy.io.fits) you need to install the astropy package')

    return


def fits_file_load(fname: str):

    grid_dict, cfg_dict = {}, {}

    if astropy_check:
        with fits.open(fname) as hdu_list:
            for i in range(1, len(hdu_list)):
                grid_dict[hdu_list[i].name] = hdu_list[i].data
    else:
        raise InnateError(f'To open ".fits" files (astropy.io.fits) you need to install the astropy package')

    return grid_dict, cfg_dict


def h5netcdf_file_save(fname: str, grid_dict: dict, common_cfg: dict, custom_cfg: dict):

    if h5netcdf_check:

        # Use the first item for the dimensions # TODO this might cause issues if grids have different size
        grid_0 = grid_dict[list(grid_dict.keys())[0]]
        data_shape = grid_0.shape

        if len(data_shape) != len(common_cfg['axes']):
            raise InnateError(f'Data set shape ({data_shape}) size is different from configuration axes'
                              f' ({common_cfg["axes"]}) size')

        with h5netcdf.File(fname, 'w') as f:

            # Unique dimensions for all the datase
            for i, axis in enumerate(common_cfg['axes']):
                f.dimensions[axis] = data_shape[i]

            # Iterate over the data grids
            for grid_name, grid in grid_dict.items():
                var = f.create_variable(grid_name, common_cfg['axes'], data=grid)
                var.attrs[REF_entry] = str(grid_name)

                # Common grid entries
                if common_cfg is not None:
                    for key, value in common_cfg.items():
                        var.attrs[key] = value

                # Local entries
                if custom_cfg is not None:
                    local_cfg = custom_cfg.get(grid_name)
                    if local_cfg is not None:
                        for key, value in local_cfg.items():
                            var.attrs[f'{grid_name}_{key}'] = value

    else:
        raise InnateError(f'To open ".nc" (h5netcdf) files you need to install the h5netcdf package')

    return


def h5netcdf_file_load(fname: str):

    grid_dict, common_cfg, local_cfg = {}, {}, {}

    if h5netcdf_check:
        with h5netcdf.File(fname, 'r') as f:

            # Loop through the grid
            for var_name in f.variables:

                # Unpack the library container
                var = f.variables[var_name]

                # Scientific data
                grid_dict[var_name] = var[...]

                # Unpack the configuration parameters
                common_cfg[var_name], local_cfg[var_name] = {}, {}
                for attr in var.attrs:

                    # Global conf
                    if not attr.startswith(var_name):
                        common_cfg[var_name][attr] = var.attrs[attr]

                    # Local conf
                    else:
                        local_cfg[var_name][attr[len(var_name)+1:]] = var.attrs[attr]

                # Convert to None if empty configuration
                if len(grid_dict[var_name]) == 0:
                    grid_dict[var_name] = None

                if len(common_cfg[var_name]) == 0:
                    common_cfg[var_name] = None

                if len(local_cfg[var_name]) == 0:
                    local_cfg[var_name] = None

    else:
        raise InnateError(f'To open ".nc" (h5netcdf) files you need to install the h5netcdf package')

    # Parse the file
    return grid_dict, common_cfg, local_cfg






# Original code from:
# https://github.com/lucidrains/siren-pytorch (MIT license)
#
# m.mieskolainen@imperial.ac.uk, 2024

import math
import torch
from torch import nn
import torch.nn.functional as F
from einops import rearrange

# helpers

def exists(val):
    return val is not None

def cast_tuple(val, repeat = 1):
    return val if isinstance(val, tuple) else ((val,) * repeat)

# sin activation

class Sine(nn.Module):
    def __init__(self, w0 = 1.):
        super().__init__()
        self.w0 = w0
    def forward(self, x):
        return torch.sin(self.w0 * x)

# siren layer

class Siren(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_out,
        w0 = 1.,
        c = 6.,
        is_first = False,
        use_bias = True,
        activation = None,
        dropout = 0.
    ):
        super().__init__()
        self.dim_in = dim_in
        self.is_first = is_first

        weight = torch.zeros(dim_out, dim_in)
        bias = torch.zeros(dim_out) if use_bias else None
        self.init_(weight, bias, c = c, w0 = w0)

        self.weight = nn.Parameter(weight)
        self.bias = nn.Parameter(bias) if use_bias else None
        self.activation = Sine(w0) if activation is None else activation
        self.dropout = nn.Dropout(dropout)

    def init_(self, weight, bias, c, w0):
        dim = self.dim_in

        w_std = (1 / dim) if self.is_first else (math.sqrt(c / dim) / w0)
        weight.uniform_(-w_std, w_std)

        if exists(bias):
            bias.uniform_(-w_std, w_std)

    def forward(self, x):
        out =  F.linear(x, self.weight, self.bias)
        out = self.activation(out)
        out = self.dropout(out)
        return out

# siren network

class SirenNet(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_hidden,
        dim_out,
        num_layers,
        w0 = 1.,
        w0_initial = 30.,
        use_bias = True,
        final_activation = None,
        dropout = 0.
    ):
        super().__init__()
        self.num_layers = num_layers
        self.dim_hidden = dim_hidden

        self.layers = nn.ModuleList([])
        for ind in range(num_layers):
            is_first = ind == 0
            layer_w0 = w0_initial if is_first else w0
            layer_dim_in = dim_in if is_first else dim_hidden

            layer = Siren(
                dim_in = layer_dim_in,
                dim_out = dim_hidden,
                w0 = layer_w0,
                use_bias = use_bias,
                is_first = is_first,
                dropout = dropout
            )

            self.layers.append(layer)

        final_activation = nn.Identity() if not exists(final_activation) else final_activation
        self.last_layer = Siren(dim_in = dim_hidden, dim_out = dim_out, w0 = w0, use_bias = use_bias, activation = final_activation)

    def forward(self, x, mods = None):
        mods = cast_tuple(mods, self.num_layers)

        for layer, mod in zip(self.layers, mods):
            x = layer(x)

            if exists(mod):
                x *= rearrange(mod, 'd -> () d')

        return self.last_layer(x)

# modulatory feed forward

class Modulator(nn.Module):
    def __init__(self, dim_in, dim_hidden, num_layers):
        super().__init__()
        self.layers = nn.ModuleList([])

        for ind in range(num_layers):
            is_first = ind == 0
            dim = dim_in if is_first else (dim_hidden + dim_in)

            self.layers.append(nn.Sequential(
                nn.Linear(dim, dim_hidden),
                nn.ReLU()
            ))

    def forward(self, z):
        x = z
        hiddens = []

        for layer in self.layers:
            x = layer(x)
            hiddens.append(x)
            x = torch.cat((x, z))

        return tuple(hiddens)

# wrapper

class SirenImageWrapper(nn.Module):
    def __init__(self, net, image_width, image_height, latent_dim = None):
        super().__init__()
        assert isinstance(net, SirenNet), 'SirenWrapper must receive a Siren network'

        self.net = net
        self.image_width = image_width
        self.image_height = image_height

        self.modulator = None
        if exists(latent_dim):
            self.modulator = Modulator(
                dim_in = latent_dim,
                dim_hidden = net.dim_hidden,
                num_layers = net.num_layers
            )

        tensors = [torch.linspace(-1, 1, steps = image_height), torch.linspace(-1, 1, steps = image_width)]
        mgrid = torch.stack(torch.meshgrid(*tensors, indexing = 'ij'), dim=-1)
        mgrid = rearrange(mgrid, 'h w c -> (h w) c')
        self.register_buffer('grid', mgrid)

    def forward(self, img = None, *, latent = None):
        modulate = exists(self.modulator)
        assert not (modulate ^ exists(latent)), 'latent vector must be only supplied if `latent_dim` was passed in on instantiation'

        mods = self.modulator(latent) if modulate else None

        coords = self.grid.clone().detach().requires_grad_()
        out = self.net(coords, mods)
        out = rearrange(out, '(h w) c -> () c h w', h = self.image_height, w = self.image_width)

        if exists(img):
            return F.mse_loss(img, out)

        return out



# Neural regression of brain image data with SIREN neural net
#
# Number of iterations required may be O(10000)
#
# m.mieskolainen@imperial.ac.uk, 2024

import sys
sys.path.append("./src/")

import torch
import torchvision.transforms as transforms
import numpy as np
from PIL import Image

import innate.torch.torch_tools as torch_tools
from innate.torch.siren import SirenNet, SirenImageWrapper


# -------------------------
# Open the input image file

img_jpeg = Image.open('examples/data/brains.jpeg')

# Define and apply the transformation to convert the image to a tensor
transform = transforms.ToTensor()
img = transform(img_jpeg).float()

# Add [dummy] batch dimension to beginning
img = img.unsqueeze(0)

torch_tools.visualize_img(img=img, savename=f'input_image.jpeg')
# -----------------------------


# -----------------------------
# Set reproducable seed first

torch_tools.set_seed(1234)

    # Create the neural net: R^{dim_in} --> R^{dim_out}
net = SirenNet(
    dim_in     = 2,     # input dimension, e.g. 2D coordinate
    dim_out    = 1,     # output dimension per coordinate, e.g. for RGB image (3)
    dim_hidden = 512,   # (hyperparam) hidden dimension
    num_layers = 4,     # (hyperparam) number of layers
    w0_initial = 30.0   # (hyperparam) init noise level for the first layer (default 30)
)

# Helper wrapper for image signals
model = SirenImageWrapper(
    net,
    image_height = img.shape[-2],
    image_width  = img.shape[-1],
)

print(f'img.shape = {img.shape} | (mean,std) = ({torch.mean(img.flatten()):0.2f}, {torch.std(img.flatten()):0.2f})')

# -----------------------------------------------
# Set the device and transfer tensors

device = torch_tools.get_device('auto')

img    = img.to(device)
model  = model.to(device)


# -----------------------------------------------
# Define the optimizer

num_iter      = 10000
lr            = 3e-4
weight_decay  = 1e-5
max_grad_norm = 1.0

optimizer = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)


# -----------------------------------------------
# Define learning rate scheduler

scheduler_param = {
  'T_0'    : 500,       # Period
  'eta_min': lr / 10,   # Minimum learning rate
  'T_mult' : 1
}

scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, **scheduler_param)
# -----------------------------------------------

modelfile  = 'siren_model.pth'
load_model = False
visualize  = False

# Load the existing model
if load_model:
    torch_tools.load_torch_model(modelfile = modelfile,
                                 model     = model,
                                 optimizer = optimizer,
                                 scheduler = scheduler)

# Training loop
# (single batch gradient descent == full image at once)
for i in range(1,num_iter+1):

    optimizer.zero_grad() #!
    loss = model(img) # MSE loss inside the wrapper
    loss.backward()
    
    # Clip gradients by norm
    torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)
        
    optimizer.step()
    scheduler.step()
    
    print(f'iter = {i} / {num_iter} | loss = {loss.item():0.3E} | lr = {scheduler.get_last_lr()[0]:0.3E}')
    
    with torch.no_grad():
        if (i % 10) == 0:
            
            # This will evaluate the model prediction implicitly
            # at the training image point grid coordinates
            pred_img = model()
            
            # Visualize
            if visualize:
                torch_tools.visualize_img(img=pred_img, savename=f'pred_image_iter_{i}.jpeg')
            
            # Save the model
            torch_tools.save_torch_model(modelfile = modelfile,
                                         model     = model,
                                         optimizer = optimizer,
                                         scheduler = scheduler)



import logging
import numpy as np
from pathlib import Path
from .io import InnateError, load_dataset
from .approximation import Approximator
from .plotting import Plotter

_logger = logging.getLogger('Innate')


def reconstruct_axes_range(data_label, axes, data_cfg, data_shape):

    axes_range = {}
    for i, dim in enumerate(axes):

        linspace_idcs = data_cfg.get(f'{dim}_range')
        if linspace_idcs is None:
            raise InnateError(
                f'The input grid "{data_label}" configuration does not include a range for the axis "{dim}"')

        axes_range[dim] = np.linspace(linspace_idcs[0], linspace_idcs[1], linspace_idcs[2])

        # Check the sizes are the same
        if axes_range[dim].size != data_shape[i]:
            _logger.warning(f'The parameter "{dim}" input dimensions ({data_shape[i]}) and '
                            f'axes dimensions ({axes_range[dim].size}) are different ')

    return axes_range


class Grid:

    """

    A class used to represent an innate Grid, which consits in a (numpy) data array and a set of configuration entries.

    In the case of interpolation techniques the user can specify the tensor library, which is used to compile the approximations.

    Parameters
    ----------
    grid_label : str
       The label for the grid.
    data_array : array-like
       The data array.
    data_cfg : dict
       Configuration dictionary for the data, containing the array dimensions information.
    tensor_library : str, optional
       The tensor library to use (default is 'pytensor').

    Attributes
    ----------
    label : str
       The label for the grid.
    description : str
       Description of the data parameter.
    data : array-like
       The data array.
    axes : list
       List of axes for the data.
    shape : tuple
       Shape of the data array.
    axes_range : list
       Range of the axes for the data.
    approx : Approximator
       An approximator instance for handling data approximations.
    plot : Plotter
       A plotter instance for visualizing the data.

    Methods
    -------
    reconstruct_axes_range(grid_label, axes, data_cfg, shape)
       Reconstructs the range of axes for the grid.

    """

    def __init__(self, grid_label, data_array, data_cfg, tensor_library='pytensor'):

        # Class attributes
        self.label = None
        self.description = None
        self.data = None
        self.axes = None
        self.shape = None
        self.axes_range = None

        # Assign attribute values
        self.label = grid_label
        self.description = data_cfg['parameter']
        self.data = data_array
        self.axes = data_cfg['axes']
        self.shape = self.data.shape
        self.axes_range = reconstruct_axes_range(grid_label, self.axes, data_cfg, self.shape)

        # Declare the function attributes treatments
        approx_techniques = data_cfg.get('approximation')
        self.approx = Approximator(self, approx_techniques, data_cfg)

        # Plotting function
        self.plot = Plotter(self)

        return


class DataSet(dict):

    """

    A class used to represent a Dataset, inheriting from Python's built-in dictionary.

    This class is initialized with data arrays and configuration parameters,
    and it unpacks these into the class dictionary.

    Parameters
    ----------
    array_dict : dict
        Dictionary containing data arrays.
    common_cfg : dict
        Dictionary containing common configuration parameters.
    local_cfg : dict
        Dictionary containing local configuration parameters.

    Attributes
    ----------
    data_labels : None or list
        Placeholder for data labels. Initialized as None.
    shape_array : None or tuple
        Placeholder for the shape of the data arrays. Initialized as None.


    """

    def __init__(self, array_dict, common_cfg, local_cfg, **kwargs):



        # Attributes
        self.data_labels = None
        self.shape_array = None

        # Unpack the individual grids into the class dictionary
        self._compile_grids(array_dict, common_cfg, local_cfg, **kwargs)

        return

    @classmethod
    def from_file(cls, fname, grid_cfg=None):

        """
        Creates a DataSet dictionarly-like object from an input file address.

        Parameters
        ----------
        fname : str
            The file address or path of the file containing the dataset.
        grid_cfg : dict, optional
            Configuration parameters for the dataset provided by the user. These values will overwrite common entries on
             the fiel configuration parameter. Default is None.

        Returns
        -------
        DataSet
            A DataSet dictionarly-like object containing the scientific arrays, the data configuration and the
            approximation techniques.

        Notes
        -----
        This method loads and parses the input data from the specified file.
        It updates the input configuration with the parameters provided by the user.

        Examples
        --------
        >>> scientific_data = DataSet.from_file('data/file.txt', grid_cfg={'param': 'value'})
        """

        # Load and parse the input data
        array_dict, common_cfg, local_cfg = load_dataset(fname)

        # Update the input configuration with the parameters from the user

        return cls(array_dict, common_cfg, local_cfg)

    def _compile_grids(self, array_dict, common_cfg, local_cfg, **kwargs):

        # Extract the keys from the data and configuration containers to create a master list for the Grid set
        common_set = set(array_dict.keys()) & set(common_cfg.keys()) & set(local_cfg.keys())
        self.data_labels = np.array(list(common_set))

        # Loop throught the dataset and create the grids
        for i, data_label in enumerate(self.data_labels):

            # Extract the Grid data
            grid_array = array_dict.get(data_label)
            data_glob_conf = common_cfg.get(data_label)
            data_local_conf = local_cfg.get(data_label)

            # Local entries update the new
            grid_cfg = data_glob_conf if data_local_conf is None else {**data_glob_conf, **data_local_conf}
            self[data_label] = Grid(data_label, grid_array, grid_cfg, **kwargs)

        return

    def extract_approximation(self, technique, label_list=None):

        label_list = label_list if label_list is not None else list(self.keys())

        approx_dict = {}
        for label in label_list:

            if technique in self[label].approx.interp.techniques:
                approx_dict[label] = getattr(self[label].approx.interp, technique)

            elif technique in self[label].approx.reg.techniques:
                approx_dict[label] = getattr(self[label].approx.reg, technique)

            else:
                _logger.critical(f'Input approximation "{technique}" is not available for dataset "{label}"')

        return approx_dict




import logging

from .pytensor import interpolation_coordinates
from .. import _setup_cfg

_logger = logging.getLogger('Innate')


class Interpolator:

    def __init__(self, grid, technique_list, tensor_library='pytensor', data_cfg=None):

        # Attributes
        self.rgi = None
        self.techniques = []

        # Confirm the data is available
        if grid.data is None:
            _logger.warning(f'The data set "{grid.label}" does not include a grid interpolation techniques cannot be used')

        # Constrain to
        algorithms = list(set(_setup_cfg['parameter_labels']['interp'].keys()) & set(technique_list))

        # Regular grid Interpolation
        if 'rgi' in algorithms:
            self.techniques.append('rgi')
            self.rgi = interpolation_coordinates(grid.data, list(grid.axes_range.values()), interp_type='point')

        return



import logging
import numpy as np
from matplotlib import pyplot as plt, rc_context, cm
from . import _setup_cfg

_logger = logging.getLogger('Innate')


class Themer:

    def __init__(self, conf, style='default', library='matplotlib'):

        # Attributes
        self.conf = None
        self.style = None
        self.base_conf = None
        self.colors = None
        self.library = None

        # Assign default
        self.conf = conf.copy()
        self.library = library
        self.set_style(style)

        return

    def fig_defaults(self, user_fig=None, fig_type=None):

        # Get plot configuration
        if fig_type is None:
            fig_conf = self.base_conf.copy()
        else:
            fig_conf = {** self.base_conf, **self.conf[self.library][fig_type]}

        # Get user configuration
        fig_conf = fig_conf if user_fig is None else {**fig_conf, **user_fig}

        return fig_conf

    def ax_defaults(self, user_ax, x_units, y_units, norm_flux, fig_type='default', **kwargs):

        # Default wavelength and flux
        if fig_type == 'plane':

            # Spectrum labels x-wavelegth, y-flux # TODO without units
            ax_cfg = {'xlabel': f'{x_units:s}', 'ylabel': f'{y_units:s}'}

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        # Spatial cubes
        elif fig_type == 'cube':

            ax_cfg = {} if user_ax is None else user_ax.copy()

            # Define the title
            if ax_cfg.get('title') is None:

                title = r'{} band'.format(kwargs['line_bg'].latex_label[0])

                line_fg = kwargs.get('line_fg')
                if line_fg is not None:
                    title = f'{title} with {line_fg.latex_label[0]} contours'

                if len(kwargs['masks_dict']) > 0:
                    title += f'\n and spatial masks at foreground'

                ax_cfg['title'] = title

            # Define x axis
            if ax_cfg.get('xlabel') is None:
                ax_cfg['xlabel'] = 'x' if kwargs['wcs'] is None else 'RA'

            # Define y axis
            if ax_cfg.get('ylabel') is None:
                ax_cfg['ylabel'] = 'y' if kwargs['wcs'] is None else 'DEC'

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        # No labels
        else:
            ax_cfg = {}

            # Update with the user configuration
            ax_cfg = ax_cfg if user_ax is None else {**ax_cfg, **user_ax}

        return ax_cfg

    def set_style(self, style=None, fig_cfg=None, colors_conf=None):

        # Set the new style
        if style is not None:
            self.style = np.atleast_1d(style)
        else:
            self.style = np.atleast_1d('default')

        # Generate the default
        self.base_conf = self.conf[self.library]['default'].copy()
        for style in self.style:
            self.base_conf = {**self.base_conf, **self.conf[self.library][style]}

        # Add the new configuration
        if fig_cfg is not None:
            self.base_conf = {**self.base_conf, **fig_cfg}

        # Set the colors
        for i_style in self.style:
            if i_style in self.conf['colors']:
                self.colors = self.conf['colors'][style].copy()

        # Add the new colors
        if colors_conf is not None:
            self.colors = {**self.colors, **colors_conf}

        if self.colors is None:
            _logger.warning(f'The input style {self.style} does not have a LiMe color database')

        return


# LiMe figure labels and color formatter
theme = Themer(_setup_cfg)

def maximize_center_fig(maximize_check=False, center_check=False):

    if maximize_check:

        # Windows maximize
        mng = plt.get_current_fig_manager()

        try:
            mng.window.showMaximized()
        except:
            try:
                mng.resize(*mng.window.maxsize())
            except:
                _logger.debug(f'Unable to maximize the window')

    if center_check:

        try:
            mngr = plt.get_current_fig_manager()
            mngr.window.setGeometry(1100, 300, mngr.canvas.width(), mngr.canvas.height())
        except:
            _logger.debug(f'Unable to center plot window')

    return


def save_close_fig_swicth(file_path=None, bbox_inches=None, fig_obj=None, maximise=False, plot_check=True):

    # By default, plot on screen unless an output address is provided
    if plot_check:
        output_fig = None

        if file_path is None:

            # Tight layout
            if bbox_inches is not None:
                plt.tight_layout()

            # Window positioning and size
            maximize_center_fig(maximise)

            # Display
            plt.show()

        else:
            plt.savefig(file_path, bbox_inches=bbox_inches)

            # Close the figure in the case of printing
            if fig_obj is not None:
                plt.close(fig_obj)

    # Return the figure for output plotting
    else:
        output_fig = fig_obj

    return output_fig


class Plotter:

    def __init__(self, grid):

        self._grid = grid

        return

    def matrix_diagnostic(self, output_address=None, num_points=15, technique='rgi', in_fig=None, fig_cfg={}, ax_cfg={},
                          maximize=False):

        """
        This function plots the 2D grid array as a surface plot where at even intervals the scatter points color represent
        the absolute divergence between the array data and the selected approximation.

        Parameters
        ----------
        output_address : str, optional
            The file path to save the output figure. If None, the figure will be displayed on the screen.
        num_points : int, optional
            The number of points per axis to test the discrepancy between the grid data and the approximation. Default is 15.
        technique : str, optional
            The interpolation technique to use. Default is 'rgi'.
        in_fig : matplotlib.figure.Figure, optional
            An existing figure to plot on. If None, a new figure will be created.
        fig_cfg : dict, optional
            Configuration dictionary for the figure settings. Default is an empty dictionary.
        ax_cfg : dict, optional
            Configuration dictionary for the axes settings. Default is an empty dictionary.
        maximize : bool, optional
            Whether to maximize the figure window. Default is False.

        Returns
        -------
        in_fig : matplotlib.figure.Figure
            The figure object containing the diagnostic plot.

        Notes
        -----
        This function prepares the data by extracting the grid, parameters, and axes ranges. It computes the test points
        and the discrepancies between the interpolated and actual data values. The results are plotted on a figure,
        highlighting areas with discrepancies below and above 1%.

        Examples
        --------
        >>> fig = obj.matrix_diagnostic(output_address='output.png', num_points=20, technique='rgi')
        >>> plt.show(fig)
        """

        # Prepare the data
        grid = self._grid.data
        params = self._grid.axes
        axes_range = self._grid.axes_range
        y_range = axes_range[params[0]]
        x_range = axes_range[params[1]]

        # Compute test points
        intvl_X, intvl_Y = (x_range.size - 1) / (num_points - 1), (y_range.size - 1) / (num_points - 1)
        idcsX = np.array([int(round(i * intvl_X)) for i in range(num_points)])
        idcsY = np.array([int(round(i * intvl_Y)) for i in range(num_points)])
        X, Y = np.meshgrid(x_range[idcsX], y_range[idcsY])
        idcmeshX, idcsmeshY = np.meshgrid(idcsX, idcsY)

        # Compute discrepancy
        emis_mesh = np.column_stack((y_range[idcsY], x_range[idcsX]))
        emis_interp = self._grid.approx.interp.rgi(emis_mesh).eval()
        emis_data = grid[idcsmeshY, idcmeshX]
        percentage_difference = np.abs(1 - emis_interp / emis_data) * 100

        # Display check for the user figures
        display_check = True if in_fig is None else False

        # Adjust the default theme
        PLT_CONF = theme.fig_defaults(fig_cfg)
        ax_cfg = ax_cfg if ax_cfg is not None else {}

        # Create and fill the figure
        with rc_context(PLT_CONF):

            if in_fig is None:
                in_fig, in_ax = plt.subplots()
            else:
                in_ax = in_fig.add_subplot()

            # Plot plane
            in_ax.imshow(grid, aspect=0.05, extent=(x_range.min(), x_range.max(), y_range.min(), y_range.max()))

            # Plot discrepancy below
            idx_interest = percentage_difference < 5
            in_ax.scatter(X[idx_interest], Y[idx_interest], c="None", edgecolors='black', linewidths=0.35, label='Error below 1%')

            if np.sum(idx_interest == False) > 0:
                scatter_err = in_ax.scatter(X[~idx_interest], Y[~idx_interest], c=percentage_difference[~idx_interest],
                                   edgecolors='black', linewidths=0.1, cmap=cm.OrRd, label='Error above 1%')

                # Color bar
                cbar = in_fig.colorbar(scatter_err)
                cbar.ax.set_ylabel('Discrepancy (%)', rotation=270, labelpad=20)

            in_ax.set_ylabel(f'{params[0]}')
            in_ax.set_xlabel(f'{params[1]}')

            in_ax.legend(loc='upper right', framealpha=1)

            in_ax.update(ax_cfg)

            # By default, plot on screen unless an output address is provided
            in_fig = save_close_fig_swicth(output_address, 'tight', in_fig, maximize, display_check)

        return in_fig


        # # Generate fitted surface points
        # matrix_edge = int(np.sqrt(te_ne_grid[0].shape[0]))
        # surface_points = func_emis(te_ne_grid, *coeffs)
        #
        # # Compare pyneb values with values from fitting
        # percentage_difference = (1 - surface_points / emis_grid) * 100
        #
        # ion, wave, latex_label = lime.label_decomposition(line_label, scalar_output=True)
        #
        # # Generate figure
        # fig = plt.figure()
        # ax = fig.add_subplot(111)


        # # Plot plane
        # ax.imshow(surface_points.reshape((matrix_edge, matrix_edge)),
        #           aspect=0.03,
        #           extent=(te_ne_grid[1].min(), te_ne_grid[1].max(), te_ne_grid[0].min(), te_ne_grid[0].max()))
        #

        # # Plot plane
        # ax.imshow(surface_points.reshape((matrix_edge, matrix_edge)),
        #           aspect=0.03,
        #           extent=(te_ne_grid[1].min(), te_ne_grid[1].max(), te_ne_grid[0].min(), te_ne_grid[0].max()))
        #
        # # Points with error below 1.0 are transparent:
        # idx_interest = percentage_difference < 0.5
        # ax.scatter(te_ne_grid[1][idx_interest], te_ne_grid[0][idx_interest], c="None", edgecolors='black',
        #            linewidths=0.35, label='Error below 1%')
        #
        # if idx_interest.sum() < emis_grid.size:
        #     # Plot grid points
        #     sc_im = ax.scatter(te_ne_grid[1][~idx_interest], te_ne_grid[0][~idx_interest],
        #                        c=percentage_difference[~idx_interest],
        #                        edgecolors='black', linewidths=0.1, cmap=cm.OrRd, label='Error above 1%')
        #
        #     # Color bar
        #     cbar = fig.colorbar(sc_im)
        #     cbar.ax.set_ylabel('Discrepancy (%)', rotation=270, labelpad=20)
        #
        # # Add labels
        # ax.set_ylabel('Temperature $(K)$', fontsize=15)
        # ax.set_xlabel('Density ($cm^{-3}$)', fontsize=15)
        # title_label = f'{latex_label}, emissivity grid\n versus parametrisation'
        # title_label = f''
        # ax.set_title(title_label, fontsize=15)
        #
        # ax.set_ylim(te_ne_grid[0].min(), te_ne_grid[0].max())
        # ax.set_xlim(te_ne_grid[1].min(), te_ne_grid[1].max())
        #
        # # Display the plot
        # ax.legend(loc='lower right', framealpha=1)
        #
        # if output_address is None:
        #     plt.show()
        # else:
        #     plt.savefig(output_address, bbox_inches='tight')





        return


import itertools
import logging
from ..io import InnateError

_logger = logging.getLogger('Innate')


try:
    import pytensor.tensor as tt
    pytensor_check = True
except ImportError:
    pytensor_check = False


def as_tensor_variable(x, dtype="float64", **kwargs):
    t = tt.as_tensor_variable(x, **kwargs)
    if dtype is None:
        return t
    return t.astype(dtype)


def interpolation_selection(grid_dict, x_range, y_range, z_range=None, interp_type='point'):

    # Container for interpolators
    interp_dict = {}

    for line, data_grid in grid_dict.items():

        # 2D Point interpolation
        if interp_type == 'point':
            interp_i = RegularGridInterpolator([x_range, y_range], data_grid[:, :, None], nout=1)

        # Line interpolation
        elif interp_type == 'axis':
            data_grid_reshape = data_grid.reshape((x_range.size, y_range.size, -1))
            interp_i = RegularGridInterpolator([x_range, y_range], data_grid_reshape)

        # 3D point
        elif interp_type == 'cube':
            interp_i = RegularGridInterpolator([x_range, y_range, z_range], data_grid)

        # Evaluate and store it
        interp_dict[line] = interp_i.evaluate

    return interp_dict


def interpolation_coordinates(data_grid, axes_range_list, z_range=None, interp_type='point'):

    # 2D Point interpolation
    if interp_type == 'point':
        interp_i = RegularGridInterpolator(axes_range_list, data_grid[:, :, None], nout=1)

    # Line interpolation
    elif interp_type == 'axis':
        data_grid_reshape = data_grid.reshape((axes_range_list[0].size, axes_range_list[1].size, -1))
        interp_i = RegularGridInterpolator(axes_range_list, data_grid_reshape)

    # 3D point
    elif interp_type == 'cube':
        interp_i = RegularGridInterpolator(axes_range_list, data_grid)

    else:
        raise InnateError(f'The interpolation type ({interp_type}) is not recognized,'
                          f' please use "point", "axis" or "cube"')

    return interp_i.evaluate


def regular_grid_interp(points, values, coords, *, fill_value=None):

    """
    Linear interpolation on a regular grid in arbitrary dimensions.

    The data must be defined on a filled regular grid, but the spacing may be
    uneven in any of the dimensions.

    Parameters
    ----------
    points : list of array-like
        A list of vectors with shapes ``(m1,), ... (mn,)``. These define the grid points in each dimension.
    values : array-like
        A tensor defining the values at each point in the grid defined by ``points``. This must have the shape ``(m1, ... mn, ..., nout)``.

    Returns
    -------
    result : array-like
        Interpolated values at the requested points.

    """

    points = [as_tensor_variable(p) for p in points]
    ndim = len(points)
    values = as_tensor_variable(values)
    coords = as_tensor_variable(coords)

    # Find where the points should be inserted
    indices = []
    norm_distances = []
    out_of_bounds = tt.zeros(coords.shape[:-1], dtype=bool)
    for n, grid in enumerate(points):
        x = coords[..., n]
        i = tt.extra_ops.searchsorted(grid, x) - 1
        out_of_bounds |= (i < 0) | (i >= grid.shape[0] - 1)
        i = tt.clip(i, 0, grid.shape[0] - 2)
        indices.append(i)
        norm_distances.append((x - grid[i]) / (grid[i + 1] - grid[i]))

    result = tt.zeros(tuple(coords.shape[:-1]) + tuple(values.shape[ndim:]))
    for edge_indices in itertools.product(*((i, i + 1) for i in indices)):
        weight = tt.ones(coords.shape[:-1])
        for ei, i, yi in zip(edge_indices, indices, norm_distances):
            weight *= tt.where(tt.eq(ei, i), 1 - yi, yi)
        result += values[edge_indices] * weight

    if fill_value is not None:
        result = tt.switch(out_of_bounds, fill_value, result)

    return result


class RegularGridInterpolator:

    """

    Linear interpolation on a regular grid in arbitrary dimensions.

    The data must be defined on a filled regular grid, but the spacing may be
    uneven in any of the dimensions.

    Parameters
    ----------
    points : list of array-like
        A list of vectors with shapes ``(m1,), ... (mn,)``. These define the grid points in each dimension.
    values : array-like
        A tensor defining the values at each point in the grid defined by ``points``. This must have the shape ``(m1, ... mn, ..., nout)``.

    Returns
    -------
    result : array-like
        Interpolated values at the requested points.

    """

    def __init__(self, points, values, fill_value=None, **kwargs):

        # Check pytensor has been installed
        if pytensor_check:
            self.ndim = len(points)
            self.points = points
            self.values = values
            self.fill_value = fill_value

        else:
            _logger.critical(f'PyTensor is not installed, this interpolation cannot be applied.')
            raise InnateError(f'Need to install PyTensor to use this function')

    def evaluate(self, t):
        """Interpolate the data

        Args:
            t: A matrix defining the coordinates where the interpolation
                should be evaluated. This must have the shape
                ``(ntest, ndim)``.
        """
        return regular_grid_interp(self.points, self.values, t, fill_value=self.fill_value)


import numpy as np
import pytest
 
from innate.regression.methods import generate_specific_function  
from innate.regression.methods import extract_coef_names  
from innate.regression.methods import extract_variables_names   
from innate.regression.methods import create_coef_dict

 


expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'
coef_names = extract_coef_names(expression)
print(coef_names)
coef_values = [1,2,3]
print(coef_values)
#X['a', 'b', 'c']
# Extract variables ending with '_range' (default suffix)
range_variables = extract_variables_names(expression)
print(range_variables) 
#['variable1_range', 'variable2_range']


print(create_coef_dict(coef_names, coef_values))
#{'a': 1, 'b': 2, 'c': 3}



expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'

# Fixed values for a, b, c
coefficients = {'a': 1, 'b': 4, 'c': 7}

# Specify the variable names that should vary
variable_names = ['variable1_range', 'variable2_range']  # These names should match usage in expression

# Create a function that varies with specified variables
@pytest.fixture 
def specific_function():
    expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'

    # Fixed values for a, b, c
    coefficients = {'a': 1, 'b': 4, 'c': 7}

    # Specify the variable names that should vary
    variable_names = ['variable1_range', 'variable2_range']
    return generate_specific_function(expression, coefficients, variable_names) 
#print( str(specific_function([10000, 20000, 30000],[10000, 20000, 30000])))
@pytest.mark.parametrize(
    "test, expected",
    [
        ([10000, 20000, 30000], [5.0, 5.107209969647869, 5.67318211637097])
    ]



)
def test_specific_function(test, expected):
    specific_function = generate_specific_function(expression, coefficients, variable_names) 
    np.testing.assert_almost_equal(
        specific_function([10000, 20000, 30000],[10000, 20000, 30000]), np.array(expected), decimal=7
    )


expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'

# Fixed values for a, b, c
coefficients = {'a': 1, 'b': 4, 'c': 7}

# Specify the variable names that should vary
variable_names = ['variable1', 'variable2']  # These names should match usage in expression

# Create a function that varies with specified variables
specific_function = generate_specific_function(expression, coefficients, variable_names)

# Example usage with arrays for each variable
variable1_values = np.array([10000, 20000, 30000])  # Example values for variable1
variable2_values = np.array([10000, 20000, 30000])  # Example values for variable2

 

expression = 'a + b / (variable1/10000.0) + c * np.log10(variable2/10000)'

# Extract variables
@pytest.mark.parametrize(
    "test, expected",
    [
        ('a + b / (variable1/10000.0) + c * np.log10(variable2/10000)', ['a', 'b', 'c'])
    ]

)
def test_extract_coef_names(test, expected):
    np.testing.assert_equal(
        extract_coef_names( test ),  expected) 
 



@pytest.mark.parametrize(
    "test, expected",
    [
        ((coef_names,coef_values),{'a': 1, 'b': 2, 'c': 3})
    ]

)
def test_create_coef_dict(test, expected):
    expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'
    coef_names = extract_coef_names(expression)
    coef_values = [1,2,3]
    np.testing.assert_equal(
       create_coef_dict(coef_names,coef_values),expected
    )

@pytest.mark.parametrize(
    "test, expected",
    [
        ('a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)',['variable1_range', 'variable2_range'])
    ]

)
def test_extract_variables_names(test, expected):
    np.testing.assert_equal(
       extract_variables_names(test ), expected
    )


# Answer should be {'a': 1, 'b': 2, 'c': 3}

# expression = 'a + b / (variable1_range/10000.0) + c * np.log10(variable2_range/10000)'

# Extract variables ending with '_range' (default suffix)
# range_variables = extract_variables_names(expression)
# print(range_variables) 
#['variable1_range', 'variable2_range']


# create_coef_dict (coef_names, coef_values)
#{'a': 1, 'b': 2, 'c': 3}


# Torch aux tools
#
# m.mieskolainen@imperial.ac.uk, 2024

import numpy as np
import torch
import random
import matplotlib.pyplot as plt


def get_device(device: str='auto'):
        
    # Force specific
    if device != 'auto':
        return torch.device(device)
    
    # Check if CUDA is available
    if torch.cuda.is_available() and device == 'auto':
        device = torch.device("cuda")
        print("get_device: CUDA device selected:", torch.cuda.get_device_name(0))
    else:
        device = torch.device("cpu")
        print("get_device: No CUDA device available, using CPU.")

    return device

def set_seed(seed: int):
    """
    Set random seed
    """
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def visualize_img(img: torch.Tensor, remove_batch_dim: bool=True,
                  savename: str=None, dpi: float=120):
    """
    Visualize torch tensor image
    """
    
    if remove_batch_dim:
        # Remove the batch dimension for visualization (from (1, C, H, W) to (C, H, W))
        img_tensor = img.squeeze(0)
    
    # Convert the tensor to a NumPy array and transpose it to (H, W, C)
    img_np = img_tensor.permute(1, 2, 0).detach().cpu().numpy()

    # Visualize the image using Matplotlib
    plt.imshow(img_np)
    plt.axis('off')  # Turn off axis numbers and ticks
    
    if savename is None:
        plt.show()
    else:
        plt.savefig(savename, bbox_inches='tight', dpi=dpi)
        print(f"visualize_img: Saved image to '{savename}'")


def load_torch_model(modelfile: str, model, optimizer=None, scheduler=None):
    """
    Load torch model (+ optimizer, + scheduler) from a file
    """
    checkpoint = torch.load(modelfile)
    
    model.load_state_dict(checkpoint['model_state_dict'])
    
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    if scheduler is not None:
        scheduler.load_state_dict(checkpoint['scheduler_state_dict'])
    
    print(f"load_torch_model: Model loaded from '{modelfile}'")


def save_torch_model(modelfile: str, model, optimizer=None, scheduler=None):
    """
    Save torch model (+ optimizer, + scheduler)
    """

    torch.save({
        'model_state_dict':     model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict() if optimizer is not None else None,
        'scheduler_state_dict': scheduler.state_dict() if scheduler is not None else None,
    }, modelfile)

    print(f"save_torch_model: Torch model saved to '{modelfile}'")




import logging

from .io import InnateError
from .interpolation.methods import Interpolator
from .regression.methods import Regressor

_logger = logging.getLogger('Innate')


class Approximator:

    def __init__(self, grid, technique_list, data_cfg=None):

        if technique_list is None:
            _logger.critical(f'The data set "{grid.label}" does not include the approximation to include the "approximation"'
                             f'key in the configurations variable or file')

        # Methodology approaches
        self.interp = Interpolator(grid, technique_list, data_cfg=data_cfg)
        self.reg = Regressor(grid, technique_list, data_cfg=data_cfg)

        return


# Neural regression of (a single) emission line data with SIREN neural net
#
# TBD. Add test-validation-test data split.
#
# TBD. Could be possible to train a single (parametric) neural net which fits
# all lines simultanouesly. Alternatively, separate model for each.
#
#
# m.mieskolainen@imperial.ac.uk, 2024

import sys
sys.path.append("./src/")

import torch
from torch.utils.data import DataLoader, TensorDataset
import numpy as np

import innate
import innate.torch.torch_tools as torch_tools
from innate.torch.siren import SirenNet


# -------------------------
# Open the input data

data_file = 'examples/data/emissivity_grids.nc'
emissivities = innate.DataSet.from_file(data_file)

# Loop through the lines
for i, line in enumerate(emissivities.data_labels):

    # Running example for just one line:
    if i == 0:
        
        grid_line = emissivities[line]

        ## Reshape data and parameter space coordinates for training
        temp_range, den_range = grid_line.axes_range['temp'], grid_line.axes_range['den']
        X, Y = np.meshgrid(temp_range, den_range)

        coords = np.column_stack((X.flatten(), Y.flatten()))
        values = grid_line.data.ravel()[:,None] # Add dim with [:,None]
        
        # Convert to pytorch tensor (float32)
        coords = torch.from_numpy(coords).float()
        values = torch.from_numpy(values).float()

        break # Take the first line

print(f'coords.shape = {coords.shape} | (mean,std) = ({torch.mean(coords), torch.std(coords)}')
print(f'values.shape = {values.shape} | (mean,std) = ({torch.mean(values), torch.std(values)}')

# -----------------------------
# Set reproducable seed first

torch_tools.set_seed(1234)

# Create the neural net: R^{dim_in} --> R^{dim_out}
model = SirenNet(
    dim_in     = 2,     # input dimension, e.g. 2D coordinate
    dim_out    = 1,     # output dimension per coordinate
    dim_hidden = 512,   # (hyperparam) hidden dimension
    num_layers = 4,     # (hyperparam) number of layers
    w0_initial = 30.0   # (hyperparam) init noise level for the first layer (default 30)
)

# -----------------------------------------------
# Set the computing device and transfer tensors

device = torch_tools.get_device('auto')

coords = coords.to(device)
values = values.to(device)
model  = model.to(device)


# -----------------------------------------------
# Define the optimizer

num_epoch     = 500
batch_size    = 1024

lr            = 1e-4
weight_decay  = 1e-5
max_grad_norm = 1.0

optimizer = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)


# -----------------------------------------------
# Define learning rate scheduler

scheduler_param = {
  'T_0'    : 100,      # Period
  'eta_min': lr / 10,  # Minimum learning rate
  'T_mult' : 1
}

scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, **scheduler_param)

# -----------------------------------------------
# Data loader for minibatching

dataset    = TensorDataset(coords, values)
dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

# -----------------------------------------------

modelfile  = 'siren_model_ngrid.pth'
load_model = False

# Load the existing model
if load_model:
    torch_tools.load_torch_model(modefile  = modelfile,
                                 model     = model,
                                 optimizer = optimizer,
                                 scheduler = scheduler)

# We use MSE loss
lossfunc = torch.nn.MSELoss(reduction='mean')

for epoch in range(1,num_epoch+1):
    
    total_loss = 0
    num_batch  = 0
    
    # Minibatch training loop
    for batch_idx, (inputs, targets) in enumerate(dataloader):
        
        optimizer.zero_grad() #!
        
        # Compute the model prediction and MSE loss
        preds = model(inputs)
        loss  = lossfunc(preds, targets)
        loss.backward()
        
        # Clip gradients by norm
        torch.nn.utils.clip_grad_norm_(model.parameters(), max_grad_norm)

        optimizer.step()
        
        # Save for output
        total_loss += loss.item()
        num_batch += 1

    scheduler.step() # Step per epoch
    
    total_loss = total_loss / num_batch

    print(f'epoch = {epoch} / {num_epoch} | loss = {total_loss:0.3E} | lr = {scheduler.get_last_lr()[0]:0.3E}')
    
    with torch.no_grad():
        if (epoch % 10) == 0:
            
            # Save the model
            torch_tools.save_torch_model(modelfile = modelfile,
                                         model     = model,
                                         optimizer = optimizer,
                                         scheduler = scheduler)

